package ru.dbotthepony.kstarbound.world.phys

import ru.dbotthepony.kstarbound.util.NotNullTwoDimensionalArray
import ru.dbotthepony.kstarbound.world.CHUNK_SHIFT
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE_FF
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ITileState
import ru.dbotthepony.kvector.vector.nint.Vector2i

class RectTileFlooderDepthFirst(
	private val tiles: NotNullTwoDimensionalArray<out ITileState>,
	private val seen: BooleanArray,
	rootx: Int,
	rooty: Int
) {
	val mins: Vector2i
	val maxs: Vector2i

	private fun filled(x: Int, y: Int): Boolean {
		if (x < 0 || x > CHUNK_SIZE_FF) {
			return false
		}

		if (y < 0 || y > CHUNK_SIZE_FF) {
			return false
		}

		return !seen[x or (y shl CHUNK_SHIFT)] && tiles[x, y].material != null
	}

	init {
		// expand wide
		var widthPositive = 1

		while (true) {
			if (!filled(rootx + widthPositive, rooty)) {
				break
			}

			widthPositive++
		}

		var widthNegative = 1

		while (true) {
			if (!filled(rootx - widthNegative, rooty)) {
				break
			}

			widthNegative++
		}

		// expand tall
		var heightPositive = 1

		while (true) {
			if (!filled(rootx, rooty + heightPositive)) {
				break
			}

			heightPositive++
		}

		var heightNegative = 1

		while (true) {
			if (!filled(rootx, rooty - heightNegative)) {
				break
			}

			heightNegative++
		}

		widthPositive -= 1
		widthNegative -= 1

		heightNegative -= 1
		heightPositive -= 1

		if (heightPositive + heightNegative > widthPositive + widthNegative) {
			// height is bigger
			// try to expand wide
			widthPositive = 0
			widthNegative = 0

			while (true) {
				var escape = false

				for (i in -heightNegative .. heightPositive) {
					if (!filled(rootx + widthPositive, rooty + i)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				widthPositive++
			}

			while (true) {
				var escape = false

				for (i in -heightNegative .. heightPositive) {
					if (!filled(rootx - widthNegative, rooty + i)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				widthNegative++
			}

			widthNegative -= 1
			widthPositive -= 1
		} else {
			// height is equal or lesser than width
			// expand high
			heightNegative = 0
			heightPositive = 0

			while (true) {
				var escape = false

				for (i in -widthNegative .. widthPositive) {
					if (!filled(rootx + i, rooty + heightPositive)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				heightPositive++
			}

			while (true) {
				var escape = false

				for (i in -widthNegative .. widthPositive) {
					if (!filled(rootx + i, rooty - heightNegative)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				heightNegative++
			}

			heightNegative -= 1
			heightPositive -= 1
		}

		mins = Vector2i(rootx - widthNegative, rooty - heightNegative)
		maxs = Vector2i(rootx + widthPositive, rooty + heightPositive)
	}

	fun markSeen() {
		for (x in mins.x .. maxs.x) {
			for (y in mins.y .. maxs.y) {
				seen[x or (y shl CHUNK_SHIFT)] = true
			}
		}
	}
}